
run
ProfMonitor subclass: 'ProfMonitorEnhanced'
  instVarNames: #( tallyBound tallyTotals elapsedCpuTime sections)
  classVars: #( PmeEntry)
  classInstVars: #()
  poolDictionaries: #[]
  inDictionary: Globals
  constraints: #[  ]
  instancesInvariant: false
  isModifiable: false
%

! Remove existing behavior from ProfMonitorEnhanced
run
ProfMonitorEnhanced removeAllMethods.
ProfMonitorEnhanced class removeAllMethods.
%

! ------------------- Class methods for ProfMonitorEnhanced
category: 'Initialization'
classmethod: ProfMonitorEnhanced
initialize
  "self initialize"
| dict |
dict := SymbolDictionary new.
PmeEntry == nil 
ifTrue: [
  PmeEntry := Object subclass: #ProfMonEntry
    instVarNames: #(#cmethod #tally #callees #rcvrClass #subtreeTallySum)
    classVars:  #()
    classInstVars:  #()
    poolDictionaries: #[ ]
    inDictionary: dict
    constraints: #[ #[#tally, SmallInteger] ]
    instancesInvariant: false isModifiable: false.
  PmeEntry compileAccessingMethodsFor: 
    #( #cmethod #tally #callees #rcvrClass subtreeTallySum) ].
%

category: 'Documentation'
classmethod: ProfMonitorEnhanced
provenance 
^ 'Proxima, a GemStone customer, requested a variation on the 
   GemStone-supplied ProfMonitor: they wanted to see a profile 
   of running code similar to the one in VAST or VW.
   Alan Purdy, a GemStone contractor reverse-engineered the interface 
   between the Gem VM and ProfMonitor, then forward-engineered this 
   ProfMonitorEnhanced class to fill the Proxima''s request. '
%

category: 'Testing'
classmethod: ProfMonitorEnhanced
runAllTests
  "self runAllTests"
self test1000.
self test10000.
self test10000withShortMaxHeight.
self test1000withObjNew.
self test5000.
self test5000Run: 5.
self test5000Times10theHardway.
self testSamplingThroughPrim.
%

category: 'Documentation'
classmethod: ProfMonitorEnhanced
simpleExample 
  " "
%

category: 'Testing'
classmethod: ProfMonitorEnhanced
test1000
  "self test1000"
| d pm |
pm := self new.
d := Dictionary new.
pm monitorBlock: [
  1 to: 1000 do: [:x | d at: x printString put: x] ].
^pm report
%

category: 'Testing'
classmethod: ProfMonitorEnhanced
test10000
  "self test10000"
| d pm |
pm := self new.
d := Dictionary new.
pm monitorBlock: [
  1 to: 10000 do: [:x | d at: x printString put: x] ].
^pm report
%

category: 'Testing'
classmethod: ProfMonitorEnhanced
test10000ObjNewViaProfMonitor
  "self test10000ObjNewViaProfMonitor"
| d pm |
pm := ProfMonitor new.
pm traceObjectCreation: true.
d := Dictionary new.
pm monitorBlock: [
  1 to: 10000 do: [:x | d at: x printString put: x] ].
^pm report
%

category: 'Testing'
classmethod: ProfMonitorEnhanced
test10000viaProfMonitor
"self test10000viaProfMonitor"
| d pm |
pm := ProfMonitor new.
d := Dictionary new.
pm monitorBlock: [
  1 to: 10000 do: [:x | d at: x printString put: x] ].
^pm report
%

category: 'Testing'
classmethod: ProfMonitorEnhanced
test10000withShortMaxHeight
  "self test10000withShortMaxHeight"
  "NOTE: when you set the maxCallStackHeight to a small number,
  the VM's samples do not include the whole stack.  We treat this
  situation (unfortunate though it is) as if there were multiple
  threads.  I sincerely hope that all call stacks are of less height."
| d pm |
pm := self new.
pm maxCallStackHeight: 8.
d := Dictionary new.
pm monitorBlock: [
  1 to: 1000 do: [:x | d at: x printString put: x] ].
^pm report
%

category: 'Testing'
classmethod: ProfMonitorEnhanced
test1000withObjNew
  "self test1000withObjNew"
| d pm |
pm := self new.
pm traceObjectCreation: true.
d := Dictionary new.
pm monitorBlock: [
  1 to: 1000 do: [:x | d at: x printString put: x] ].
^pm report
%

category: 'Testing'
classmethod: ProfMonitorEnhanced
test5000
  "self test5000"
| d pm |
pm := self new.
d := Dictionary new.
pm monitorBlock: [
  1 to: 5000 do: [:x | d at: x printString put: x] ].
^pm report
%

category: 'Testing'
classmethod: ProfMonitorEnhanced
test5000Run: nTimes
  "self test5000Run: 10"
| d pm  |
pm := self new.
d := Dictionary new.
pm monitorBlock: [
  1 to: 5000 do: [:x | d at: x printString put: x] ] repeatTimes: nTimes.
^pm report
%

category: 'Testing'
classmethod: ProfMonitorEnhanced
test5000Times10theHardway
  "self test5000Times10theHardway"
| d pm  |
pm := self new.
d := Dictionary new.
pm monitorBlock: [
pm pauseMonitoring.
10 timesRepeat: [
  System abortTransaction.
  pm resumeMonitoring.
  1 to: 5000 do: [:x | d at: x printString put: x].
    pm pauseMonitoring ] ].
^pm report
%

category: 'Testing'
classmethod: ProfMonitorEnhanced
testSamplingThroughPrim
  "During testing, the client ran into a special case:  Consider what 
   happens if you have an identity set with an index.  The add: message 
   must invoke some Smalltalk code to update the indexes to reflect the 
   added entry.  It does this in a weird and tricky manner of (apparently) 
   failing the primitive, then updating the relevant indexes from within 
   the an exception handler.  But, I may have misunderstood it.  The add: 
   definitely calls a primitive, but the primitive appears to actually put 
   something on the call stack, invoking _updateIndexesForAdditionOf:logging:.  
   But, the primitive itself appears to call other entries in a normal manner, 
   and at one point, triggering an exception.  So, perhaps the exception is 
   irrelevant, and the operative cause of the failure of this code is the 
   existence of a nil on the sample list.  I don't know how to trigger one of 
   these without using indexes on an identity set."
"self testSamplingThroughPrim"
| d pm  setClass elementClass |
setClass := System myUserProfile objectNamed: #SetContract.  
"At least, compile in all environments"
setClass == nil 
ifTrue: [
  ^self error: 'You are not in correct environment for a test'].
elementClass := System myUserProfile objectNamed: #Contract.
pm := self new.
d := setClass new.
d createIdentityIndexOn: 'adresse'.
pm monitorBlock: [
  1 to: 500 do: [:x |
    d add: elementClass new. ] ].
^pm report
%

! ------------------- Instance methods for ProfMonitorEnhanced
category: 'private report generation'
method: ProfMonitorEnhanced
cgrHeaderTo: rpt
| lf  elapsedTime missingSamples |
lf := Character lf.
rpt addAll: 'STATISTICAL CALL GRAPH RESULTS'; add: lf ; add: lf;
    addAll: 'Elapsed CPU time:    ';
    addAll: (elapsedCpuTime asString); addAll: ' ms'; add: lf.
rpt addAll: 'Monitoring interval: '; 
    addAll: interval asString; addAll: ' ms'; add: lf .
"Get rid of this because it should always be set to 20"
"rpt addAll: 'Max call stack height: ';
     addAll: sampleDepth asString; add: lf."   
rpt addAll: 'Total samples (hits): ';
    addAll: tallyTotals asString; add: lf.
missingSamples := ((elapsedCpuTime / interval) asInteger) - tallyTotals.
rpt addAll: 'Samples in VM or user-excluded areas: ';
    addAll: missingSamples asString; addAll: '  ('; 
    addAll: (missingSamples * interval) asString; addAll: ' msec.)';
    add: lf; add: lf.
%

category: 'private report generation'
method: ProfMonitorEnhanced
cgrSubgraphFor: pmeEntry to: rpt level: treeLevel
  "At each level of the graph, put out a line for this pmEntry of the 
   form: treeLevelTabs [ '[] in '] ReceiverClass [ (implementer's class if 
   different ) ] methodName
   percentage of total tally in this branch (summed from below)
   percentage of hits in this method
   number of hits in this method 
   Sum of time spent in this method or below
   If counting object creations, the number of objects created below.
   If a primitive, time spent in the primitive."
| method pctJustHere pctEnclosed rcvrClass timeBase |
treeLevel timesRepeat: [rpt addAll: '     '].
method := pmeEntry cmethod.
rcvrClass := pmeEntry rcvrClass.
(rcvrClass isKindOf: BlockClosure class) 
ifTrue: [  
  "It is a block of some kind"
  rpt add: $[; addAll: (rcvrClass name); addAll: '] in '.
  "Was rcvrClass _cost asString"
  method selector == nil 
  ifTrue: [ "a doit "
    rpt addAll: (method inClass == nil 
      ifTrue: [''] 
      ifFalse: [method inClass name]).
    rpt addAll: '>>'; addAll: '(a doit)' ] 
  ifFalse: [
    rpt addAll: (method inClass name); addAll: '>>'; add: (method selector) ] ] 
ifFalse: [
  method == true 
  ifTrue: [  "Its an object creation, tally is number of objects."
    rpt addAll: '(PRIMITIVE)  Created '; addAll: (pmeEntry tally asString);
        addAll: ' instances of '; addAll: (rcvrClass name); add: Character lf.
    ^self ] 
  ifFalse: [   
    method selector == nil 
    ifTrue: [  "it is a Do it"
      rpt addAll: (rcvrClass name); addAll: '>>'; addAll: '(a doit)' ] 
    ifFalse: [   
      method _class == GsMethod 
      ifTrue: [ "It is a standard method"
        rpt addAll: rcvrClass name.
        rcvrClass == method inClass 
        ifFalse: [
          rpt add: $(; addAll: (method inClass name); add: $)].
          rpt addAll: '>>'; addAll: method selector ] 
      ifFalse: [
        method == #GCI 
        ifTrue: [   "It is a call to a gci primitive.. "
          rpt addAll: (rcvrClass name); addAll: '>>#Gci prim' ] 
        ifFalse: [  " It is an error..."
          self error: 'Unknown key in a call stack trace.' ] ] ] ] ].
"Put % of total of enclosed calls; % of hits in this context"
"990927 Guy Leblanc wanted percentages to be based on elapsedCpuTime"
timeBase := elapsedCpuTime // interval.
pctJustHere := (pmeEntry tally asFloat * 100.0 / timeBase) asStringUsingFormat: #(-3 1 false).
pctEnclosed := (pmeEntry subtreeTallySum asFloat  * 100.0 / timeBase) asStringUsingFormat: #(-3 1 false).
rpt addAll: ' ('; addAll: pctEnclosed; addAll: '%,'; addAll: pctJustHere; addAll: '%) ['; 
    addAll: pmeEntry tally asString;
    addAll: (pmeEntry tally == 1 ifTrue: [' hit]'] ifFalse: [' hits]']); 
    add: Character lf.
"Now recur to all the kids.."
pmeEntry callees == nil ifTrue: [^self].
pmeEntry callees do: [:anotherPmeEntry | 
  self cgrSubgraphFor: anotherPmeEntry to: rpt level: treeLevel + 1 ].
%

category: 'private report generation'
method: ProfMonitorEnhanced
cgrSumSubgraphTallies: pmEntry
  "Sum the tallies of the subgraph subTotals.
   set up your own, return that value."
| subtreeSum |
pmEntry cmethod == true ifTrue: [^0].
subtreeSum := pmEntry tally.        "Add yourself too."
pmEntry callees == nil ifFalse: [
pmEntry callees do: [:calledNode |
  subtreeSum := subtreeSum + (self cgrSumSubgraphTallies: calledNode) ]        ].
pmEntry subtreeTallySum: subtreeSum.
^subtreeSum
%

category: 'private report generation'
method: ProfMonitorEnhanced
cgrTallyTotalsFor: meths
| total aMeth |
total := 0.
meths do: [:each |
  aMeth := each cmethod.
  (aMeth _class == GsMethod _or:[ aMeth == #GCI ]) 
  ifTrue:[
    total := total + each tally ] ].
^total
%

category: 'Initializers'
method: ProfMonitorEnhanced
defaultCallStackHeight
^20
%

category: 'private sample processing'
method: ProfMonitorEnhanced
explainStatsFile
  "Each stats file is a function of depth variable, with the default being 2.
   I override this (in initNoFile) to 20.  This number indicates how 
   many contexts on the stack to represent whenever the timer fires at each
   sampling interval. (the depth must be at least 2).  If the number is
   smaller than the contexts on the call stack, then it takes depth
   contexts starting from the active context, clipping off the root.
   When processing the stats file, I essentially must turn it around,
   and build a rooted tree of pmEntries.  But, the pmEntry is basically 
   backwards, it links an entry to its sender, not an entry to the methods it 
   invokes.  To get around this, I'm using the sender's field (a collection) in the 
   other direction, it is really the message-rcvrClass pair represented by 
   each entry in the senders collection (each entry in this is also a pmEntry.
   When we build it, if we truncate the sample's list of contexts because 
   the depth is too short, we keep an unrooted section in the dictionary to 
   deal with this problem.  So the dictionary passed around is:
     keyed on selector-rcvrClass,
     value is aPmEntry
   When we get another sample, we lookup the last entry in the dictionary, 
   then move forward through the message-rcvrClass pairs to get to the active 
   context.  Once there, we update the tally on that leaf."
%

category: 'Reporting'
method: ProfMonitorEnhanced
gatherResults
  "Analyze the receiver's file of sampling data and store the results of 
   the analysis in the results instance variable of the receiver."
results := self _processStatsFile.
%

category: 'Initializers'
method: ProfMonitorEnhanced
initialize
  "Override my superclass to initialize the call stack height (aka sampleDepth)
   to the default 50."
super initialize.
self maxCallStackHeight: self defaultCallStackHeight.
elapsedCpuTime := 0.
"In some weird cases, the file is not being deleted properly.  Slam it again."
self removeFile.
self _createFile: self class newProfileFileName.
file close.
%

category: 'Initializers'
method: ProfMonitorEnhanced
maxCallStackHeight: maxHeight
  "The ProfMonitor gear in the VM sets a timer (see instVar interval).
   Whenever the timer fires (it fires every 'interval' milliseconds),
   it examines the call stack at that moment, and for each context,
   records the compiled method and the receiver's class in a 'Sample
   array' starting with the active context.  For whatever reason, the
   ProfMonitor implimentors wanted to use fixed-size arrays for each
   sample.  The sampleDepth instVar communicates to the VM the
   size of that sample array.  So if you set maxHeight too low, a deep
   call stack's sample will truncate the top - i.e., from the first context
   of a process.  If you set it too high, it will write nils in the
   higher sample array slots, and take up some extra time and space in the 
   sample file.  I see little reason not to set it to a number larger 
   than what you think is your max process stack height."
   "Unfortunately, the VM used to restrict this sampleDepth number to  2 .. 20."
(maxHeight between: 2 and: 50) 
ifFalse: [self error: 'Call stack height must be in range of 2..20'].
        sampleDepth := maxHeight
%

category: 'Monitoring'
method: ProfMonitorEnhanced
monitorBlock: aBlock
  "Similar to System | millisecondsToRun:, this method starts 
   profiling, executes the block, and terminates profiling."
self _openFileAppend.
elapsedCpuTime := 0.
self resumeMonitoring.
aBlock value.
self pauseMonitoring.
file close.
self gatherResults; removeFile; _removeSampleArray .
%

category: 'Monitoring'
method: ProfMonitorEnhanced
monitorBlock: aBlock repeatTimes: nTimes
  "Similar to System | millisecondsToRun:, this method starts 
   profiling, executes the block, and terminates profiling."
| intervalMsec |
self _openFileAppend.
elapsedCpuTime := 0.
nTimes timesRepeat: [
  self resumeMonitoring.
  aBlock value.
  self pauseMonitoring ].
file close.
self gatherResults; removeFile; _removeSampleArray
%

category: 'Monitoring'
method: ProfMonitorEnhanced
pauseMonitoring
  "Stops monitoring.  Override the super to accumulate elapsedCpuTime.
   We do this because you may turn on and off monitoring multiple times, 
   accumulating samples across when it is turned off."
| thisIntervalmsec |
endTime := System _readClock.
self _zeroArgPrim: 2  . "stop monitoring in the virtual machine"
thisIntervalmsec := 
  (endTime >= startTime 
  ifTrue: [endTime - startTime]
  ifFalse: [ endTime + (16rffffffff - startTime)]).
elapsedCpuTime := elapsedCpuTime + thisIntervalmsec.
%

category: 'Reporting'
method: ProfMonitorEnhanced
reportDownTo: tally
  "Adds report spliting out things by call graph. "
| rpt lf aReport |
results == nil ifTrue: [^'No report data available'].
rpt := String new.
lf := Character lf.
tallyBound := tally.
aReport := self _callGraphReport.
rpt add: lf; addAll: aReport;  add: lf ; add: lf.
aReport size: 0. 
^rpt
%

category: 'Monitoring'
method: ProfMonitorEnhanced
resumeMonitoring
  "Starts monitoring."
self _zeroArgPrim: 1 .  "enable monitoring in the virtual machine"
startTime := System _readClock.
%

category: 'Monitoring'
method: ProfMonitorEnhanced
traceObjectCreation: aBoolean
  "Enable (aBoolean == true) or disable profiling of object creation. 
   The state change will take effect on the next invocation of 
   ProfMonitor>>startMonitoring for the receiver."
aBoolean _validateClass: Boolean.
traceObjCreation := aBoolean.
%

category: 'private report generation'
method: ProfMonitorEnhanced
_callGraphReport
| callGraph nameStr rpt unreportedInfo pmEntries  |
rpt := String new.
callGraph := results.
tallyTotals := 0.
callGraph valuesDo: [:pmeEntry |
  tallyTotals := tallyTotals + (self cgrSumSubgraphTallies: pmeEntry) ].
self cgrHeaderTo: rpt.
callGraph do: [:pmeEntry | self cgrSubgraphFor: pmeEntry to: rpt level: 0].
^rpt
%

category: 'private sample processing'
method: ProfMonitorEnhanced
_processStatsFile
  "File = { Section }*  "
  "Usually, there is only one section in the sample file.  However, if 
   you run multiple samples via the #monitorBlock:repeatTimes: method,
   it needs to accumulate multiple sections. "
| callGraph rawArray rawStream  |
sections := 0.
callGraph := Dictionary new .
rawArray := self _readSampleFile.
rawStream := ReadStream on: rawArray.
[rawStream atEnd ] whileFalse: [
  self _readSection: rawStream to: callGraph.
  sections := sections + 1 ].
^callGraph
%

category: 'private sample processing'
method: ProfMonitorEnhanced
_readNormalSampleOf: depth from: rawStream to: callGraph
  "MethodContext = aGsMethod rcvrClass
   PrimContext = #GCI rcvrClass  "
| sample subGraph methodAndRcvr foundEntry |
sample := self _readSampleOf: depth from: rawStream.
subGraph := callGraph.
1 to: sample size  do: [:idx |        
  "Want to use sample size downTo: 1 do: = , but that is difficult to step through"
  methodAndRcvr := sample at: (sample size + 1 - idx).
  ((foundEntry := subGraph at: methodAndRcvr otherwise: nil) == nil)
  ifTrue: [
    foundEntry := PmeEntry new
      cmethod: (methodAndRcvr key);
      rcvrClass: (methodAndRcvr value);
      tally: 0;
      callees: (Dictionary new);
      yourself.
    subGraph at: methodAndRcvr put: foundEntry ].
  subGraph := foundEntry callees.
  idx == sample size 
  ifTrue: [
    foundEntry tally: (foundEntry tally + 1)] ].
%

category: 'private sample processing'
method: ProfMonitorEnhanced
_readSampleOf: depth from: rawStream
  "Return an array of pairs, with depth contexts, each an association.
   The array is cut off at the first pair of nil entries. Then we turn it 
   around to put the root of the call stack first. "
| sample max aMeth rcvrClass |
sample := Array new: depth.
max := 0.
1 to: depth do: [:i |
  aMeth := rawStream next.
  rcvrClass := rawStream next.
  (rcvrClass == nil _and: [aMeth == nil]) 
  ifFalse: [  "a prim = has nil rcvrClass, but perhaps a non-nil meth"
    max := max + 1.
    sample at: max put: (aMeth->rcvrClass) ] ].
    sample size: max.
^sample
%

category: 'private sample processing'
method: ProfMonitorEnhanced
_readSampleOf: depth from: rawStream to: callGraph
  "Sample = ActiveContext { Context }  [ nil* ]   ..for total of 
   depth entries ActiveContext = MethodContext | ObjNewContext | PrimContext | 
   BadContext Context = MethodContext | NilContext | BadContext
   MethodContext = aGsMethod rcvrClass
   ObjNewContext = true newObjClass
   PrimContext = #GCI rcvrClass
   NilContext = nil nil - past the top of the call stack.
   BadContext = anyOtherObj anyOtherObj - This is in error
   A sample represents the call chain starting from the active context. 
   If the depth is too small, then the rightmost context will not be the root 
   context."
| first |
rawStream atEnd ifTrue: [^false].
first := rawStream peek.
first == SystemRepository 
ifTrue: [^false].
(first == true _or: [first _class == GsMethod _or: [ first == #GCI]]) 
ifTrue: [
  self _readNormalSampleOf: depth from: rawStream to: callGraph.
  ^true ].
first == nil 
ifTrue: [
  self _readSampleOf: depth from: rawStream.        "Then discard it"
  ^true ] 
ifFalse: [
  self halt: '** Not one of the keys I was expecting ' ].
%

category: 'private sample processing'
method: ProfMonitorEnhanced
 _readSection: rawStream to: callGraph
  "Section =3D aRepository  DepthInteger  { Sample }"
|  depth  |
rawStream next == SystemRepository 
ifFalse: [self halt: '** Expected a Repository'].
depth := rawStream next.
depth _isSmallInteger 
ifFalse: [self halt: '** Expected integer depth.'].
[self _readSampleOf: depth from: rawStream to: callGraph] 
whileTrue: [].
%

